# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import print_function

import itertools
import os
import shutil
import time
from collections import namedtuple
from contextlib import contextmanager

import pkg_resources
from pkg_resources import safe_name

from .common import safe_mkdir
from .fetcher import Fetcher
from .interpreter import PythonInterpreter
from .iterator import Iterator, IteratorInterface
from .orderedset import OrderedSet
from .package import Package, distribution_compatible
from .platforms import Platform
from .resolvable import ResolvableRequirement, resolvables_from_iterable
from .resolver_options import ResolverOptionsBuilder
from .tracer import TRACER
from .util import DistributionHelper


@contextmanager
def patched_packing_env(env):
  """Monkey patch packaging.markers.default_environment"""
  old_env = pkg_resources.packaging.markers.default_environment
  new_env = lambda: env
  pkg_resources._vendor.packaging.markers.default_environment = new_env
  try:
    yield
  finally:
    pkg_resources._vendor.packaging.markers.default_environment = old_env


class Untranslateable(Exception):
  pass


class Unsatisfiable(Exception):
  pass


class StaticIterator(IteratorInterface):
  """An iterator that iterates over a static list of packages."""

  def __init__(self, packages, allow_prereleases=None):
    self._packages = packages
    self._allow_prereleases = allow_prereleases

  def iter(self, req):
    for package in self._packages:
      if package.satisfies(req, allow_prereleases=self._allow_prereleases):
        yield package


class _ResolvedPackages(namedtuple('_ResolvedPackages',
                                   'resolvable packages parent constraint_only')):

  @classmethod
  def empty(cls):
    return cls(None, OrderedSet(), None, False)

  def merge(self, other):
    if other.resolvable is None:
      return _ResolvedPackages(self.resolvable, self.packages, self.parent, self.constraint_only)
    return _ResolvedPackages(
        self.resolvable,
        self.packages & other.packages,
        self.parent,
        self.constraint_only and other.constraint_only)


class _ResolvableSet(object):
  @classmethod
  def normalize(cls, name):
    return safe_name(name).lower()

  def __init__(self, tuples=None):
    # A list of _ResolvedPackages
    self.__tuples = tuples or []

  def _collapse(self):
    # Collapse all resolvables by name along with the intersection of all compatible packages.
    # If the set of compatible packages is the empty set, then we cannot satisfy all the
    # specifications for a particular name (e.g. "setuptools==2.2 setuptools>4".)
    #
    # We need to return the resolvable since it carries its own network context and configuration
    # regarding package precedence.  This is arbitrary -- we could just as easily say "last
    # resolvable wins" but it seems highly unlikely this will materially affect anybody
    # adversely but could be the source of subtle resolution quirks.
    resolvables = {}
    for resolved_packages in self.__tuples:
      key = self.normalize(resolved_packages.resolvable.name)
      previous = resolvables.get(key, _ResolvedPackages.empty())
      if previous.resolvable is None:
        resolvables[key] = resolved_packages
      else:
        resolvables[key] = previous.merge(resolved_packages)
    return resolvables

  def _synthesize_parents(self, name):
    def render_resolvable(resolved_packages):
      return '%s%s' % (
          str(resolved_packages.resolvable),
          '(from: %s)' % resolved_packages.parent if resolved_packages.parent else '')
    return ', '.join(
        render_resolvable(resolved_packages) for resolved_packages in self.__tuples
        if self.normalize(resolved_packages.resolvable.name) == self.normalize(name))

  def _check(self):
    # Check whether or not the resolvables in this set are satisfiable, raise an exception if not.
    for name, resolved_packages in self._collapse().items():
      if not resolved_packages.packages:
        raise Unsatisfiable(
          'Could not satisfy all requirements for %s:\n    %s' % (
            resolved_packages.resolvable,
            self._synthesize_parents(name)
          )
        )

  def merge(self, resolvable, packages, parent=None):
    """Add a resolvable and its resolved packages."""
    self.__tuples.append(_ResolvedPackages(resolvable, OrderedSet(packages),
                                           parent, resolvable.is_constraint))
    self._check()

  def get(self, name):
    """Get the set of compatible packages given a resolvable name."""
    resolvable, packages, parent, constraint_only = self._collapse().get(
        self.normalize(name), _ResolvedPackages.empty())
    return packages

  def packages(self):
    """Return a snapshot of resolvable => compatible packages set from the resolvable set."""
    return list(self._collapse().values())

  def extras(self, name):
    return set.union(
        *[set(tup.resolvable.extras()) for tup in self.__tuples
          if self.normalize(tup.resolvable.name) == self.normalize(name)])

  def replace_built(self, built_packages):
    """Return a copy of this resolvable set but with built packages.

    :param dict built_packages: A mapping from a resolved package to its locally built package.
    :returns: A new resolvable set with built package replacements made.
    """
    def map_packages(resolved_packages):
      packages = OrderedSet(built_packages.get(p, p) for p in resolved_packages.packages)
      return _ResolvedPackages(resolved_packages.resolvable, packages,
                               resolved_packages.parent, resolved_packages.constraint_only)

    return _ResolvableSet([map_packages(rp) for rp in self.__tuples])


class ResolvedDistribution(namedtuple('ResolvedDistribution', 'requirement distribution')):
  """A requirement and the resolved distribution that satisfies it."""


class Resolver(object):
  """Interface for resolving resolvable entities into python packages."""

  class Error(Exception): pass

  @staticmethod
  def _maybe_expand_platform(interpreter, platform=None):
    # Expands `platform` if it is 'current' and abbreviated.
    #
    # IE: If we're on linux and handed a platform of `None`, 'current', or 'linux_x86_64', we expand
    # the platform to an extended platform matching the given interpreter's abi info, eg:
    # 'linux_x86_64-cp-27-cp27mu'.

    cur_plat = Platform.current()
    def expand_platform():
      expanded_platform = Platform(platform=cur_plat.platform,
                                   impl=interpreter.identity.abbr_impl,
                                   version=interpreter.identity.impl_ver,
                                   abi=interpreter.identity.abi_tag)
      TRACER.log("""
Modifying given platform of {given_platform!r}:
Using the current platform of {current_platform!r}
Under current interpreter {current_interpreter!r}

To match given interpreter {given_interpreter!r}.

Calculated platform: {calculated_platform!r}""".format(
        given_platform=platform,
        current_platform=cur_plat,
        current_interpreter=PythonInterpreter.get(),
        given_interpreter=interpreter,
        calculated_platform=expanded_platform),
        V=9
      )
      return expanded_platform

    if platform in (None, 'current'):
      # Always expand the default local (abbreviated) platform to the given interpreter.
      return expand_platform()
    else:
      given_platform = Platform.create(platform)
      if given_platform.is_extended:
        # Always respect an explicit extended platform.
        return given_platform
      elif given_platform.platform != cur_plat.platform:
        # IE: Say we're on OSX and platform was 'linux-x86_64'; we can't expand a non-local
        # platform so we leave as-is.
        return given_platform
      else:
        # IE: Say we're on 64 bit linux and platform was 'linux-x86_64'; ie: the abbreviated local
        # platform.
        return expand_platform()

  def __init__(self, allow_prereleases=None, interpreter=None, platform=None, use_manylinux=None):
    self._interpreter = interpreter or PythonInterpreter.get()
    self._platform = self._maybe_expand_platform(self._interpreter, platform)
    self._allow_prereleases = allow_prereleases
    platform_name = self._platform.platform
    self._target_interpreter_env = self._interpreter.identity.pkg_resources_env(platform_name)
    self._supported_tags = self._platform.supported_tags(
      self._interpreter,
      use_manylinux
    )
    TRACER.log(
      'R: tags for %r x %r -> %s' % (self._platform, self._interpreter, self._supported_tags),
      V=9
    )

  def filter_packages_by_supported_tags(self, packages, supported_tags=None):
    return [
      package for package in packages
      if package.compatible(supported_tags or self._supported_tags)
    ]

  def package_iterator(self, resolvable, existing=None):
    if existing:
      existing = resolvable.compatible(
        StaticIterator(existing, allow_prereleases=self._allow_prereleases))
    else:
      existing = resolvable.packages()
    return self.filter_packages_by_supported_tags(existing)

  def build(self, package, options):
    context = options.get_context()
    translator = options.get_translator(self._interpreter, self._supported_tags)
    with TRACER.timed('Fetching %s' % package.url, V=2):
      local_package = Package.from_href(context.fetch(package))
    if local_package is None:
      raise Untranslateable('Could not fetch package %s' % package)
    with TRACER.timed('Translating %s into distribution' % local_package.local_path, V=2):
      dist = translator.translate(local_package)
    if dist is None:
      raise Untranslateable('Package %s is not translateable by %s' % (package, translator))
    if not distribution_compatible(dist, self._supported_tags):
      raise Untranslateable(
        'Could not get distribution for %s on platform %s.' % (package, self._platform))
    return dist

  def is_resolvable_in_target_interpreter_env(self, resolvable):
    if not isinstance(resolvable, ResolvableRequirement):
      return True
    elif resolvable.requirement.marker is None:
      return True
    else:
      return resolvable.requirement.marker.evaluate(environment=self._target_interpreter_env)

  def resolve(self, resolvables, resolvable_set=None):
    resolvables = [(resolvable, None) for resolvable in resolvables
                   if self.is_resolvable_in_target_interpreter_env(resolvable)]
    resolvable_set = resolvable_set or _ResolvableSet()
    processed_resolvables = set()
    processed_packages = {}
    distributions = {}

    while resolvables:
      while resolvables:
        resolvable, parent = resolvables.pop(0)
        if resolvable in processed_resolvables:
          continue
        packages = self.package_iterator(resolvable, existing=resolvable_set.get(resolvable.name))

        resolvable_set.merge(resolvable, packages, parent)
        processed_resolvables.add(resolvable)

      built_packages = {}
      for resolvable, packages, parent, constraint_only in resolvable_set.packages():
        if constraint_only:
          continue
        assert len(packages) > 0, 'ResolvableSet.packages(%s) should not be empty' % resolvable
        package = next(iter(packages))
        if resolvable.name in processed_packages:
          if package == processed_packages[resolvable.name]:
            continue
        if package not in distributions:
          dist = self.build(package, resolvable.options)
          built_package = Package.from_href(dist.location)
          built_packages[package] = built_package
          distributions[built_package] = dist
          package = built_package

        distribution = distributions[package]
        processed_packages[resolvable.name] = package
        new_parent = '%s->%s' % (parent, resolvable) if parent else str(resolvable)
        # We patch packaging.markers.default_environment here so we find optional reqs for the
        # platform we're building the PEX for, rather than the one we're on.
        with patched_packing_env(self._target_interpreter_env):
          resolvables.extend(
            (ResolvableRequirement(req, resolvable.options), new_parent) for req in
            distribution.requires(extras=resolvable_set.extras(resolvable.name))
          )
      resolvable_set = resolvable_set.replace_built(built_packages)

    # We may have built multiple distributions depending upon if we found transitive dependencies
    # for the same package. But ultimately, resolvable_set.packages() contains the correct version
    # for all packages. So loop through it and only return the package version in
    # resolvable_set.packages() that is found in distributions.
    dists = []
    # No point in proceeding if distributions is empty
    if not distributions:
      return dists

    for resolvable, packages, parent, constraint_only in resolvable_set.packages():
      if constraint_only:
        continue
      assert len(packages) > 0, 'ResolvableSet.packages(%s) should not be empty' % resolvable
      package = next(iter(packages))
      distribution = distributions[package]
      if isinstance(resolvable, ResolvableRequirement):
        requirement = resolvable.requirement
      else:
        requirement = distribution.as_requirement()
      dists.append(ResolvedDistribution(requirement=requirement,
                                        distribution=distribution))
    return dists


class CachingResolver(Resolver):
  """A package resolver implementing a package cache."""

  @classmethod
  def filter_packages_by_ttl(cls, packages, ttl, now=None):
    now = now if now is not None else time.time()
    return [package for package in packages
        if package.remote or package.local and (now - os.path.getmtime(package.local_path)) < ttl]

  def __init__(self, cache, cache_ttl, *args, **kw):
    self.__cache = cache
    self.__cache_ttl = cache_ttl
    safe_mkdir(self.__cache)
    super(CachingResolver, self).__init__(*args, **kw)

  # Short-circuiting package iterator.
  def package_iterator(self, resolvable, existing=None):
    iterator = Iterator(fetchers=[Fetcher([self.__cache])],
                        allow_prereleases=self._allow_prereleases)
    packages = self.filter_packages_by_supported_tags(resolvable.compatible(iterator))

    if packages and self.__cache_ttl:
      packages = self.filter_packages_by_ttl(packages, self.__cache_ttl)

    return itertools.chain(
      packages,
      super(CachingResolver, self).package_iterator(resolvable, existing=existing)
    )

  # Caching sandwich.
  def build(self, package, options):
    # cache package locally
    if package.remote:
      package = Package.from_href(options.get_context().fetch(package, into=self.__cache))
      os.utime(package.local_path, None)

    # build into distribution
    dist = super(CachingResolver, self).build(package, options)

    # if distribution is not in cache, copy
    target = os.path.join(self.__cache, os.path.basename(dist.location))
    if not os.path.exists(target):
      shutil.copyfile(dist.location, target + '~')
      os.rename(target + '~', target)
    os.utime(target, None)

    return DistributionHelper.distribution_from_path(target)


def platform_to_tags(platform, interpreter):
  """Splits a "platform" like linux_x86_64-36-cp-cp36m into its components.

  If a simple platform without hyphens is specified, we will fall back to using
  the current interpreter's tags.
  """
  if platform.count('-') >= 3:
    tags = platform.rsplit('-', 3)
  else:
    tags = [platform, interpreter.identity.impl_ver,
            interpreter.identity.abbr_impl, interpreter.identity.abi_tag]
  tags[0] = tags[0].replace('.', '_').replace('-', '_')
  return tags


def resolve(requirements,
            fetchers=None,
            interpreter=None,
            platform=None,
            context=None,
            precedence=None,
            cache=None,
            cache_ttl=None,
            allow_prereleases=None,
            use_manylinux=None):
  """Produce all distributions needed to (recursively) meet `requirements`

  :param requirements: An iterator of Requirement-like things, either
    :class:`pkg_resources.Requirement` objects or requirement strings.
  :keyword fetchers: (optional) A list of :class:`Fetcher` objects for locating packages.  If
    unspecified, the default is to look for packages on PyPI.
  :keyword interpreter: (optional) A :class:`PythonInterpreter` object to use for building
    distributions and for testing distribution compatibility.
  :keyword versions: (optional) a list of string versions, of the form ["33", "32"],
    or None. The first version will be assumed to support our ABI.
  :keyword platform: (optional) specify the exact platform you want valid
    tags for, or None. If None, use the local system platform.
  :keyword impl: (optional) specify the exact implementation you want valid
    tags for, or None. If None, use the local interpreter impl.
  :keyword abi: (optional) specify the exact abi you want valid
    tags for, or None. If None, use the local interpreter abi.
  :keyword context: (optional) A :class:`Context` object to use for network access.  If
    unspecified, the resolver will attempt to use the best available network context.
  :keyword precedence: (optional) An ordered list of allowable :class:`Package` classes
    to be used for producing distributions.  For example, if precedence is supplied as
    ``(WheelPackage, SourcePackage)``, wheels will be preferred over building from source, and
    eggs will not be used at all.  If ``(WheelPackage, EggPackage)`` is suppplied, both wheels and
    eggs will be used, but the resolver will not resort to building anything from source.
  :keyword cache: (optional) A directory to use to cache distributions locally.
  :keyword cache_ttl: (optional integer in seconds) If specified, consider non-exact matches when
    resolving requirements.  For example, if ``setuptools==2.2`` is specified and setuptools 2.2 is
    available in the cache, it will always be used.  However, if a non-exact requirement such as
    ``setuptools>=2,<3`` is specified and there exists a setuptools distribution newer than
    cache_ttl seconds that satisfies the requirement, then it will be used.  If the distribution
    is older than cache_ttl seconds, it will be ignored.  If ``cache_ttl`` is not specified,
    resolving inexact requirements will always result in making network calls through the
    ``context``.
  :keyword allow_prereleases: (optional) Include pre-release and development versions.  If
    unspecified only stable versions will be resolved, unless explicitly included.
  :keyword use_manylinux: (optional) Whether or not to use manylinux for linux resolves.
  :returns: List of :class:`ResolvedDistribution` instances meeting ``requirements``.
  :raises Unsatisfiable: If ``requirements`` is not transitively satisfiable.
  :raises Untranslateable: If no compatible distributions could be acquired for
    a particular requirement.

  This method improves upon the setuptools dependency resolution algorithm by maintaining sets of
  all compatible distributions encountered for each requirement rather than the single best
  distribution encountered for each requirement.  This prevents situations where ``tornado`` and
  ``tornado==2.0`` could be treated as incompatible with each other because the "best
  distribution" when encountering ``tornado`` was tornado 3.0.  Instead, ``resolve`` maintains the
  set of compatible distributions for each requirement as it is encountered, and iteratively filters
  the set.  If the set of distributions ever becomes empty, then ``Unsatisfiable`` is raised.

  .. versionchanged:: 0.8
    A number of keywords were added to make requirement resolution slightly easier to configure.
    The optional ``obtainer`` keyword was replaced by ``fetchers``, ``translator``, ``context``,
    ``threads``, ``precedence``, ``cache`` and ``cache_ttl``, also all optional keywords.

  .. versionchanged:: 1.0
    The ``translator`` and ``threads`` keywords have been removed.  The choice of threading
    policy is now implicit.  The choice of translation policy is dictated by ``precedence``
    directly.

  .. versionchanged:: 1.0
    ``resolver`` is now just a wrapper around the :class:`Resolver` and :class:`CachingResolver`
    classes.

  .. versionchanged:: 1.5.0
    The ``pkg_blacklist``  has been removed and the return type changed to a list of
    :class:`ResolvedDistribution`.
  """

  builder = ResolverOptionsBuilder(fetchers=fetchers,
                                   allow_prereleases=allow_prereleases,
                                   use_manylinux=use_manylinux,
                                   precedence=precedence,
                                   context=context)

  if cache:
    resolver = CachingResolver(cache,
                               cache_ttl,
                               allow_prereleases=allow_prereleases,
                               use_manylinux=use_manylinux,
                               interpreter=interpreter,
                               platform=platform)
  else:
    resolver = Resolver(allow_prereleases=allow_prereleases,
                        use_manylinux=use_manylinux,
                        interpreter=interpreter,
                        platform=platform)

  return resolver.resolve(resolvables_from_iterable(requirements, builder))


def resolve_multi(requirements,
                  fetchers=None,
                  interpreters=None,
                  platforms=None,
                  context=None,
                  precedence=None,
                  cache=None,
                  cache_ttl=None,
                  allow_prereleases=None,
                  use_manylinux=None):
  """A generator function that produces all distributions needed to meet `requirements`
  for multiple interpreters and/or platforms.

  :param requirements: An iterator of Requirement-like things, either
    :class:`pkg_resources.Requirement` objects or requirement strings.
  :keyword fetchers: (optional) A list of :class:`Fetcher` objects for locating packages.  If
    unspecified, the default is to look for packages on PyPI.
  :keyword interpreters: (optional) An iterable of :class:`PythonInterpreter` objects to use
    for building distributions and for testing distribution compatibility.
  :keyword platforms: (optional) An iterable of PEP425-compatible platform strings to use for
    filtering compatible distributions.  If unspecified, the current platform is used, as
    determined by `Platform.current()`.
  :keyword context: (optional) A :class:`Context` object to use for network access.  If
    unspecified, the resolver will attempt to use the best available network context.
  :keyword precedence: (optional) An ordered list of allowable :class:`Package` classes
    to be used for producing distributions.  For example, if precedence is supplied as
    ``(WheelPackage, SourcePackage)``, wheels will be preferred over building from source, and
    eggs will not be used at all.  If ``(WheelPackage, EggPackage)`` is suppplied, both wheels and
    eggs will be used, but the resolver will not resort to building anything from source.
  :keyword cache: (optional) A directory to use to cache distributions locally.
  :keyword cache_ttl: (optional integer in seconds) If specified, consider non-exact matches when
    resolving requirements.  For example, if ``setuptools==2.2`` is specified and setuptools 2.2 is
    available in the cache, it will always be used.  However, if a non-exact requirement such as
    ``setuptools>=2,<3`` is specified and there exists a setuptools distribution newer than
    cache_ttl seconds that satisfies the requirement, then it will be used.  If the distribution
    is older than cache_ttl seconds, it will be ignored.  If ``cache_ttl`` is not specified,
    resolving inexact requirements will always result in making network calls through the
    ``context``.
  :keyword allow_prereleases: (optional) Include pre-release and development versions.  If
    unspecified only stable versions will be resolved, unless explicitly included.
  :yields: All :class:`ResolvedDistribution` instances meeting ``requirements``.
  :raises Unsatisfiable: If ``requirements`` is not transitively satisfiable.
  :raises Untranslateable: If no compatible distributions could be acquired for
    a particular requirement.
  """

  interpreters = interpreters or [PythonInterpreter.get()]
  platforms = platforms or ['current']

  seen = set()
  for interpreter in interpreters:
    for platform in platforms:
      for resolvable in resolve(requirements,
                                fetchers,
                                interpreter,
                                platform,
                                context,
                                precedence,
                                cache,
                                cache_ttl,
                                allow_prereleases,
                                use_manylinux=use_manylinux):
        if resolvable not in seen:
          seen.add(resolvable)
          yield resolvable



# This file was forked from the pip project master branch on 2016/12/05

"""Generate and work with PEP 425 Compatibility Tags."""

from __future__ import absolute_import

import distutils.util
import logging
import platform
import re
import sys
import sysconfig
import warnings

from .glibc import have_compatible_glibc

logger = logging.getLogger(__name__)

_OSX_ARCH_PAT = re.compile(r'(.+)_(\d+)_(\d+)_(.+)')


def get_config_var(var):
  try:
    return sysconfig.get_config_var(var)
  except IOError as e:  # Issue #1074
    warnings.warn("{0}".format(e), RuntimeWarning)
    return None


def get_abbr_impl():
  """Return abbreviated implementation name."""
  if hasattr(sys, 'pypy_version_info'):
    pyimpl = 'pp'
  elif sys.platform.startswith('java'):
    pyimpl = 'jy'
  elif sys.platform == 'cli':
    pyimpl = 'ip'
  else:
    pyimpl = 'cp'
  return pyimpl


def get_impl_ver():
  """Return implementation version."""
  impl_ver = get_config_var("py_version_nodot")
  if not impl_ver or get_abbr_impl() == 'pp':
    impl_ver = ''.join(map(str, get_impl_version_info()))
  return impl_ver


def get_impl_version_info():
  """Return sys.version_info-like tuple for use in decrementing the minor
  version."""
  if get_abbr_impl() == 'pp':
    # as per https://github.com/pypa/pip/issues/2882
    return (sys.version_info[0], sys.pypy_version_info.major,
            sys.pypy_version_info.minor)
  else:
    return sys.version_info[0], sys.version_info[1]


def get_impl_tag():
  """
  Returns the Tag for this specific implementation.
  """
  return "{0}{1}".format(get_abbr_impl(), get_impl_ver())


def get_flag(var, fallback, expected=True, warn=True):
  """Use a fallback method for determining SOABI flags if the needed config
  var is unset or unavailable."""
  val = get_config_var(var)
  if val is None:
    if warn:
      logger.debug("Config variable '%s' is unset, Python ABI tag may "
                   "be incorrect", var)
    return fallback()
  return val == expected


def get_abi_tag():
  """Return the ABI tag based on SOABI (if available) or emulate SOABI
  (CPython 2, PyPy)."""
  soabi = get_config_var('SOABI')
  impl = get_abbr_impl()
  if not soabi and impl in ('cp', 'pp') and hasattr(sys, 'maxunicode'):
    d = ''
    m = ''
    u = ''
    if get_flag('Py_DEBUG',
                lambda: hasattr(sys, 'gettotalrefcount'),
                warn=(impl == 'cp')):
      d = 'd'
    if get_flag('WITH_PYMALLOC',
                lambda: impl == 'cp',
                warn=(impl == 'cp')):
      m = 'm'
    if (get_flag('Py_UNICODE_SIZE',
                 lambda: sys.maxunicode == 0x10ffff,
                 expected=4,
                 warn=(impl == 'cp' and sys.version_info < (3, 3))) and
        sys.version_info < (3, 3)):
      u = 'u'
    abi = '%s%s%s%s%s' % (impl, get_impl_ver(), d, m, u)
  elif soabi and soabi.startswith('cpython-'):
    abi = 'cp' + soabi.split('-')[1]
  elif soabi:
    abi = soabi.replace('.', '_').replace('-', '_')
  else:
    abi = None
  return abi


def _is_running_32bit():
  return sys.maxsize == 2147483647


def get_platform():
  """Return our platform name 'win32', 'linux_x86_64'"""
  if sys.platform == 'darwin':
    # distutils.util.get_platform() returns the release based on the value
    # of MACOSX_DEPLOYMENT_TARGET on which Python was built, which may
    # be significantly older than the user's current machine.
    release, _, machine = platform.mac_ver()
    split_ver = release.split('.')

    if machine == 'x86_64' and _is_running_32bit():
      machine = 'i386'
    elif machine == 'ppc64' and _is_running_32bit():
      machine = 'ppc'

    return 'macosx_{0}_{1}_{2}'.format(split_ver[0], split_ver[1], machine)

  # XXX remove distutils dependency
  result = distutils.util.get_platform().replace('.', '_').replace('-', '_')
  if result == 'linux_x86_64' and _is_running_32bit():
    # 32 bit Python program (running on a 64 bit Linux): pip should only
    # install and run 32 bit compiled extensions in that case.
    result = 'linux_i686'

  return result


def is_manylinux1_compatible():
  # Only Linux, and only x86-64 / i686
  if get_platform() not in ('linux_x86_64', 'linux_i686'):
    return False

  # Check for presence of _manylinux module
  try:
    import _manylinux
    return bool(_manylinux.manylinux1_compatible)
  except (ImportError, AttributeError):
    # Fall through to heuristic check below
    pass

  # Check glibc version. CentOS 5 uses glibc 2.5.
  return have_compatible_glibc(2, 5)


def get_darwin_arches(major, minor, machine):
  """Return a list of supported arches (including group arches) for
  the given major, minor and machine architecture of an macOS machine.
  """
  arches = []

  def _supports_arch(major, minor, arch):
    # Looking at the application support for macOS versions in the chart
    # provided by https://en.wikipedia.org/wiki/OS_X#Versions it appears
    # our timeline looks roughly like:
    #
    # 10.0 - Introduces ppc support.
    # 10.4 - Introduces ppc64, i386, and x86_64 support, however the ppc64
    #    and x86_64 support is CLI only, and cannot be used for GUI
    #    applications.
    # 10.5 - Extends ppc64 and x86_64 support to cover GUI applications.
    # 10.6 - Drops support for ppc64
    # 10.7 - Drops support for ppc
    #
    # Note: The above information is taken from the "Application support"
    #     column in the chart not the "Processor support" since I believe
    #     that we care about what instruction sets an application can use
    #     not which processors the OS supports.
    if arch == 'ppc':
      return (major, minor) <= (10, 5)
    if arch == 'ppc64':
      return (major, minor) == (10, 5)
    if arch == 'i386':
      return (major, minor) >= (10, 4)
    if arch == 'x86_64':
      return (major, minor) >= (10, 4)
    if arch in groups:
      for garch in groups_dict[arch]:
        if _supports_arch(major, minor, garch):
          return True
    return False

  groups = ('fat', 'intel', 'fat64', 'fat32')
  groups_dict = {'fat': ('i386', 'ppc'),
                 'intel': ('x86_64', 'i386'),
                 'fat64': ('x86_64', 'ppc64'),
                 'fat32': ('x86_64', 'i386', 'ppc')}

  if _supports_arch(major, minor, machine):
    arches.append(machine)

  for garch in groups:
    if machine in groups_dict[garch] and _supports_arch(major, minor, garch):
      arches.append(garch)

  arches.append('universal')

  return arches


def _gen_all_abis(impl, version):
  def tmpl_abi(impl, version, suffix):
    return ''.join((impl, version, suffix))
  yield tmpl_abi(impl, version, 'd')
  yield tmpl_abi(impl, version, 'dm')
  yield tmpl_abi(impl, version, 'dmu')
  yield tmpl_abi(impl, version, 'm')
  yield tmpl_abi(impl, version, 'mu')
  yield tmpl_abi(impl, version, 'u')


def get_supported_for_any_abi(version=None, noarch=False, platform=None, impl=None,
                              force_manylinux=False):
  """Generates supported tags for unspecified ABI types to support more intuitive cross-platform
     resolution."""
  unique_tags = {
    tag for abi in _gen_all_abis(impl, version)
    for tag in get_supported(version=version,
                             noarch=noarch,
                             platform=platform,
                             impl=impl,
                             abi=abi,
                             force_manylinux=force_manylinux)
  }
  return list(unique_tags)


def get_supported(version=None, noarch=False, platform=None, impl=None, abi=None,
                  force_manylinux=False):
  """Return a list of supported tags for each version specified in
  `version`.

  :param version: string version (e.g., "33", "32") or None.
    If None, use local system Python version.
  :param platform: specify the exact platform you want valid
    tags for, or None. If None, use the local system platform.
  :param impl: specify the exact implementation you want valid
    tags for, or None. If None, use the local interpreter impl.
  :param abi: specify the exact abi you want valid
    tags for, or None. If None, use the local interpreter abi.
  :param force_manylinux: Whether or not to force manylinux support. This is useful
                          when resolving for different target platform than current.
  """
  supported = []

  # Versions must be given with respect to the preference
  if version is None:
    versions = []
    version_info = get_impl_version_info()
    major = version_info[:-1]
    # Support all previous minor Python versions.
    for minor in range(version_info[-1], -1, -1):
      versions.append(''.join(map(str, major + (minor,))))
  else:
    versions = [version]

  impl = impl or get_abbr_impl()

  abis = []

  abi = abi or get_abi_tag()
  if abi:
    abis[0:0] = [abi]

  abi3s = set()
  import imp
  for suffix in imp.get_suffixes():
    if suffix[0].startswith('.abi'):
      abi3s.add(suffix[0].split('.', 2)[1])

  abis.extend(sorted(list(abi3s)))

  abis.append('none')

  if not noarch:
    arch = platform or get_platform()
    if arch.startswith('macosx'):
      # support macosx-10.6-intel on macosx-10.9-x86_64
      match = _OSX_ARCH_PAT.match(arch)
      if match:
        name, major, minor, actual_arch = match.groups()
        tpl = '{0}_{1}_%i_%s'.format(name, major)
        arches = []
        for m in reversed(range(int(minor) + 1)):
          for a in get_darwin_arches(int(major), m, actual_arch):
            arches.append(tpl % (m, a))
      else:
        # arch pattern didn't match (?!)
        arches = [arch]
    elif (
      (platform is None and is_manylinux1_compatible()) or
      # N.B. Here we work around the fact that `is_manylinux1_compatible()` expects
      # to be running on the target platform being built for with a feature flag approach.
      (arch.startswith('linux') and force_manylinux)
    ):
      arches = [arch.replace('linux', 'manylinux1'), arch]
    else:
      arches = [arch]

    # Current version, current API (built specifically for our Python):
    for abi in abis:
      for arch in arches:
        supported.append(('%s%s' % (impl, versions[0]), abi, arch))

    # abi3 modules compatible with older version of Python
    for version in versions[1:]:
      # abi3 was introduced in Python 3.2
      if version in ('31', '30'):
        break
      for abi in abi3s:   # empty set if not Python 3
        for arch in arches:
          supported.append(('%s%s' % (impl, version), abi, arch))

    # Has binaries, does not use the Python API:
    for arch in arches:
      supported.append(('py%s' % (versions[0][0]), 'none', arch))

  # No abi / arch, but requires our implementation:
  supported.append(('%s%s' % (impl, versions[0]), 'none', 'any'))
  # Tagged specifically as being cross-version compatible
  # (with just the major version specified)
  supported.append(('%s%s' % (impl, versions[0][0]), 'none', 'any'))

  # No abi / arch, generic Python
  for i, version in enumerate(versions):
    supported.append(('py%s' % (version,), 'none', 'any'))
    if i == 0:
      supported.append(('py%s' % (version[0]), 'none', 'any'))

  return supported



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os
import traceback
from abc import abstractmethod

from .archiver import Archiver
from .common import chmod_plus_w, safe_copy, safe_mkdtemp, safe_rmtree
from .compatibility import AbstractClass
from .installer import WheelInstaller
from .interpreter import PythonInterpreter
from .package import EggPackage, Package, SourcePackage, WheelPackage
from .pep425tags import get_supported
from .tracer import TRACER
from .util import DistributionHelper


class TranslatorBase(AbstractClass):
  """
    Translate a link into a distribution.
  """

  @abstractmethod
  def translate(self, link, into=None):
    pass


class ChainedTranslator(TranslatorBase):
  """
    Glue a sequence of Translators together in priority order.  The first Translator to resolve a
    requirement wins.
  """

  def __init__(self, *translators):
    self._translators = list(filter(None, translators))
    for tx in self._translators:
      if not isinstance(tx, TranslatorBase):
        raise ValueError('Expected a sequence of translators, got %s instead.' % type(tx))

  def translate(self, package, into=None):
    for tx in self._translators:
      dist = tx.translate(package, into=into)
      if dist:
        return dist

  def __str__(self):
    return 'ChainedTranslator(%s)' % (
        ', '.join((tx.__class__.__name__ for tx in self._translators)))


class SourceTranslator(TranslatorBase):
  @classmethod
  def run_2to3(cls, path):
    from lib2to3.refactor import get_fixers_from_package, RefactoringTool
    rt = RefactoringTool(get_fixers_from_package('lib2to3.fixes'))
    with TRACER.timed('Translating %s' % path):
      for root, dirs, files in os.walk(path):
        for fn in files:
          full_fn = os.path.join(root, fn)
          if full_fn.endswith('.py'):
            with TRACER.timed('%s' % fn, V=3):
              try:
                chmod_plus_w(full_fn)
                rt.refactor_file(full_fn, write=True)
              except IOError:
                TRACER.log('Failed to translate %s' % fn)
                TRACER.log(traceback.format_exc())

  def __init__(self,
               interpreter=PythonInterpreter.get(),
               supported_tags=None,
               use_2to3=False,
               installer_impl=WheelInstaller):
    self._supported_tags = supported_tags or get_supported()
    self._interpreter = interpreter
    self._installer_impl = installer_impl
    self._use_2to3 = use_2to3

  def translate(self, package, into=None):
    """From a SourcePackage, translate to a binary distribution."""
    if not isinstance(package, SourcePackage):
      return None
    if not package.local:
      raise ValueError('SourceTranslator cannot translate remote packages.')

    installer = None
    version = self._interpreter.version
    unpack_path = Archiver.unpack(package.local_path)
    into = into or safe_mkdtemp()

    try:
      if self._use_2to3 and version >= (3,):
        with TRACER.timed('Translating 2->3 %s' % package.name):
          self.run_2to3(unpack_path)
      installer = self._installer_impl(
          unpack_path,
          interpreter=self._interpreter,
          strict=(package.name not in ('distribute', 'setuptools')))
      with TRACER.timed('Packaging %s' % package.name):
        try:
          dist_path = installer.bdist()
        except self._installer_impl.InstallFailure as e:
          TRACER.log('Failed to install package at %s: %s' % (unpack_path, e))
          return None
        target_path = os.path.join(into, os.path.basename(dist_path))
        safe_copy(dist_path, target_path)
        target_package = Package.from_href(target_path)
        if not target_package:
          TRACER.log('Target path %s does not look like a Package.' % target_path)
          return None
        if not target_package.compatible(self._supported_tags):
          TRACER.log('Target package %s is not compatible with %s' % (
              target_package, self._supported_tags))
          return None
        return DistributionHelper.distribution_from_path(target_path)
    except Exception as e:
      TRACER.log('Failed to translate %s' % package)
      TRACER.log(traceback.format_exc())
    finally:
      if installer:
        installer.cleanup()
      if unpack_path:
        safe_rmtree(unpack_path)


class BinaryTranslator(TranslatorBase):
  def __init__(self,
               package_type,
               supported_tags=None):
    self._package_type = package_type
    self._supported_tags = supported_tags or get_supported()

  def translate(self, package, into=None):
    """From a binary package, translate to a local binary distribution."""
    if not package.local:
      raise ValueError('BinaryTranslator cannot translate remote packages.')
    if not isinstance(package, self._package_type):
      return None
    if not package.compatible(self._supported_tags):
      TRACER.log('Target package %s is not compatible with %s' % (
          package, self._supported_tags))
      return None
    into = into or safe_mkdtemp()
    target_path = os.path.join(into, package.filename)
    safe_copy(package.local_path, target_path)
    return DistributionHelper.distribution_from_path(target_path)


class EggTranslator(BinaryTranslator):
  def __init__(self, **kw):
    super(EggTranslator, self).__init__(EggPackage, **kw)


class WheelTranslator(BinaryTranslator):
  def __init__(self, **kw):
    super(WheelTranslator, self).__init__(WheelPackage, **kw)


class Translator(object):
  @staticmethod
  def default(interpreter=None, supported_tags=None):
    interpreter = interpreter or PythonInterpreter.get()
    supported_tags = supported_tags or get_supported()
    whl_translator = WheelTranslator(supported_tags=supported_tags)
    egg_translator = EggTranslator(supported_tags=supported_tags)
    source_translator = SourceTranslator(interpreter=interpreter)
    return ChainedTranslator(whl_translator, egg_translator, source_translator)



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os

from pkg_resources import EGG_NAME, parse_version, safe_name, safe_version

from .archiver import Archiver
from .base import maybe_requirement
from .link import Link
from .pep425tags import get_supported
from .util import Memoizer


class Package(Link):
  """Base class for named Python binary packages (e.g. source, egg, wheel)."""

  class Error(Exception): pass
  class InvalidPackage(Error): pass

  # The registry of concrete implementations
  _REGISTRY = set()

  # The cache of packages that we have already constructed.
  _HREF_TO_PACKAGE_CACHE = Memoizer()

  @classmethod
  def register(cls, package_type):
    """Register a concrete implementation of a Package to be recognized by pex."""
    if not issubclass(package_type, cls):
      raise TypeError('package_type must be a subclass of Package.')
    cls._REGISTRY.add(package_type)

  @classmethod
  def from_href(cls, href, **kw):
    """Convert from a url to Package.

    :param href: The url to parse
    :type href: string
    :returns: A Package object if a valid concrete implementation exists, otherwise None.
    """
    package = cls._HREF_TO_PACKAGE_CACHE.get(href)
    if package is not None:
      return package
    link_href = Link.wrap(href)
    for package_type in cls._REGISTRY:
      try:
        package = package_type(link_href.url, **kw)
        break
      except package_type.InvalidPackage:
        continue
    if package is not None:
      cls._HREF_TO_PACKAGE_CACHE.store(href, package)
    return package

  @property
  def name(self):
    return NotImplementedError

  @property
  def raw_version(self):
    return NotImplementedError

  @property
  def version(self):
    return parse_version(self.raw_version)

  def satisfies(self, requirement, allow_prereleases=None):
    """Determine whether this package matches the requirement.

    :param requirement: The requirement to compare this Package against
    :type requirement: string or :class:`pkg_resources.Requirement`
    :param Optional[bool] allow_prereleases: Whether to allow prereleases to satisfy
      the `requirement`.
    :returns: True if the package matches the requirement, otherwise False
    """
    requirement = maybe_requirement(requirement)
    link_name = safe_name(self.name).lower()
    if link_name != requirement.key:
      return False

    # NB: If we upgrade to setuptools>=34 the SpecifierSet used here (requirement.specifier) will
    # come from a non-vendored `packaging` package and pex's bootstrap code in `PEXBuilder` will
    # need an update.
    return requirement.specifier.contains(self.raw_version, prereleases=allow_prereleases)

  def compatible(self, supported_tags):
    """Is this package compatible with the given tag set?

    :param supported_tags: A list of tags that is supported by the target
                           interpeter, as generated by
                           :func:`pex.pep425tags.get_supported`.
    :type supported_tags: list of 3-tuples
    """
    raise NotImplementedError()


class SourcePackage(Package):
  """A Package representing an uncompiled/unbuilt source distribution."""

  @classmethod
  def split_fragment(cls, fragment):
    """A heuristic used to split a string into version name/fragment:

       >>> SourcePackage.split_fragment('pysolr-2.1.0-beta')
       ('pysolr', '2.1.0-beta')
       >>> SourcePackage.split_fragment('cElementTree-1.0.5-20051216')
       ('cElementTree', '1.0.5-20051216')
       >>> SourcePackage.split_fragment('pil-1.1.7b1-20090412')
       ('pil', '1.1.7b1-20090412')
       >>> SourcePackage.split_fragment('django-plugin-2-2.3')
       ('django-plugin-2', '2.3')
    """
    def likely_version_component(enumerated_fragment):
      return sum(bool(v and v[0].isdigit()) for v in enumerated_fragment[1].split('.'))
    fragments = fragment.split('-')
    if len(fragments) == 1:
      return fragment, ''
    max_index, _ = max(enumerate(fragments), key=likely_version_component)
    return '-'.join(fragments[0:max_index]), '-'.join(fragments[max_index:])

  def __init__(self, url, **kw):
    super(SourcePackage, self).__init__(url, **kw)

    ext = Archiver.get_extension(self.filename)
    if ext is None:
      raise self.InvalidPackage('%s is not a recognized archive format.' % self.filename)

    fragment = self.filename[:-len(ext)]
    self._name, self._raw_version = self.split_fragment(fragment)

  @property
  def name(self):
    return safe_name(self._name)

  @property
  def raw_version(self):
    return safe_version(self._raw_version)

  # SourcePackages are always compatible as they can be translated to a distribution.
  def compatible(self, supported_tags):
    return True


class BinaryPackage(Package):
  """A Package representing a binary distribution which can be e.g. platform specific."""

  def __init__(self, *args, **kwargs):
    super(BinaryPackage, self).__init__(*args, **kwargs)
    self._supported_tags = None

  @property
  def supported_tags(self):
    if not self._supported_tags:
      self._supported_tags = frozenset(self._iter_tags())
    return self._supported_tags

  def _iter_tags(self):
    raise NotImplementedError()

  def compatible(self, supported_tags):
    """Is this package compatible with the given tag set?

    :param supported_tags: A list of tags that is supported by the target
                           interpeter, as generated by
                           :func:`pex.pep425tags.get_supported`.
    :type supported_tags: list of 3-tuples
    """
    return not set(supported_tags).isdisjoint(self.supported_tags)


class EggPackage(BinaryPackage):
  """A Package representing a built egg."""

  def __init__(self, url, **kw):
    super(EggPackage, self).__init__(url, **kw)
    filename, ext = os.path.splitext(self.filename)
    if ext.lower() != '.egg':
      raise self.InvalidPackage('Not an egg: %s' % filename)
    matcher = EGG_NAME(filename)
    if not matcher:
      raise self.InvalidPackage('Could not match egg: %s' % filename)

    self._name, self._raw_version, self._py_version, self._platform = matcher.group(
        'name', 'ver', 'pyver', 'plat'
    )

    if self._raw_version is None or self._py_version is None:
      raise self.InvalidPackage('url with .egg extension but bad name: %s' % url)

  def __hash__(self):
    return hash((self.name, self.version, self.py_version, self.platform))

  @property
  def name(self):
    return safe_name(self._name)

  @property
  def raw_version(self):
    return safe_version(self._raw_version)

  @property
  def py_version(self):
    return self._py_version

  @property
  def platform(self):
    return self._platform

  def _iter_tags(self):
    abi_tag = 'none'
    tag_platform = (self._platform or 'any').replace('-', '_').replace('.', '_')
    suffix_maj = self._py_version.split('.', 1)[0]
    suffix_min = self._py_version.replace('.', '')

    for prefix in ('py', 'cp'):
      for suffix in (suffix_maj, suffix_min):
        impl = ''.join((prefix, suffix))
        yield (impl, abi_tag, tag_platform)

    # Work around PyPy versions being weird in pep425tags.get_supported
    if self.py_version == '2.7':
      yield ('pp2', abi_tag, tag_platform)
    elif self.py_version == '3.2':
      yield ('pp321', abi_tag, tag_platform)
    elif self.py_version == '3.3':
      # N.B. PyPy 3.3 maps to `pp352` because of PyPy versioning.
      # see e.g. http://doc.pypy.org/en/latest/release-pypy3.3-v5.2-alpha1.html
      yield ('pp352', abi_tag, tag_platform)
    elif self.py_version == '3.5':
      yield ('pp357', abi_tag, tag_platform)


class WheelPackage(BinaryPackage):
  """A Package representing a built wheel."""

  def __init__(self, url, **kw):
    super(WheelPackage, self).__init__(url, **kw)
    filename, ext = os.path.splitext(self.filename)
    if ext.lower() != '.whl':
      raise self.InvalidPackage('Not a wheel: %s' % filename)
    try:
      self._name, self._raw_version, self._py_tag, self._abi_tag, self._arch_tag = (
          filename.rsplit('-', 4))
    except ValueError:
      raise self.InvalidPackage('Wheel filename malformed.')
    # See https://github.com/pypa/pip/issues/1150 for why this is unavoidable.
    self._name = self._name.replace('_', '-')
    self._raw_version = self._raw_version.replace('_', '-')

  @property
  def name(self):
    return self._name

  @property
  def raw_version(self):
    return self._raw_version

  def _iter_tags(self):
    for py in self._py_tag.split('.'):
      for abi in self._abi_tag.split('.'):
        for arch in self._arch_tag.split('.'):
          yield (py, abi, arch)

  def __eq__(self, other):
    return (self._name == other._name and
            self._raw_version == other._raw_version and
            self._py_tag == other._py_tag and
            self._abi_tag == other._abi_tag and
            self._arch_tag == other._arch_tag)

  def __hash__(self):
    return hash((self._name, self._raw_version, self._py_tag, self._abi_tag, self._arch_tag))


Package.register(SourcePackage)
Package.register(EggPackage)
Package.register(WheelPackage)


def distribution_compatible(dist, supported_tags=None):
  """Is this distribution compatible with the given interpreter/platform combination?

  :param supported_tags: A list of tag tuples specifying which tags are supported
    by the platform in question.
  :returns: True if the distribution is compatible, False if it is unrecognized or incompatible.
  """
  if supported_tags is None:
    supported_tags = get_supported()
  package = Package.from_href(dist.location)
  if not package:
    return False
  return package.compatible(supported_tags)



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from pkg_resources import Requirement

from pex.crawler import Crawler
from pex.fetcher import PyPIFetcher
from pex.iterator import Iterator
from pex.package import SourcePackage

try:
  from unittest import mock
except ImportError:
  import mock


def test_empty_iteration():
  crawler_mock = mock.create_autospec(Crawler, spec_set=True)
  crawler_mock.crawl.return_value = []
  iterator = Iterator(crawler=crawler_mock)

  assert list(iterator.iter(Requirement.parse('foo'))) == []
  assert len(crawler_mock.crawl.mock_calls) == 1
  _, args, kwargs = crawler_mock.crawl.mock_calls[0]
  assert list(args[0]) == list(PyPIFetcher().urls(Requirement.parse('foo')))
  assert kwargs == {'follow_links': False}


def assert_iteration(all_versions, *expected_versions, **iterator_kwargs):
  def package_url(version):
    return 'https://pypi.org/packages/source/p/pex/pex-%s.tar.gz' % version

  urls = [package_url(v) for v in all_versions]
  crawler_mock = mock.create_autospec(Crawler, spec_set=True)
  crawler_mock.crawl.return_value = urls
  iterator = Iterator(crawler=crawler_mock, follow_links=True, **iterator_kwargs)

  assert list(iterator.iter(Requirement.parse('pex'))) == [SourcePackage(package_url(v))
                                                           for v in expected_versions]
  assert len(crawler_mock.crawl.mock_calls) == 1
  _, _, kwargs = crawler_mock.crawl.mock_calls[0]
  assert kwargs == {'follow_links': True}


def test_iteration_with_return():
  assert_iteration(['0.8.6', '0.8.6b1'], '0.8.6')  # Iterator should exclude prereleases by default.
  assert_iteration(['0.8.6', '0.8.6b1'], '0.8.6', allow_prereleases=False)
  assert_iteration(['0.8.6', '0.8.6b1'], '0.8.6', '0.8.6b1', allow_prereleases=True)



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess

import pytest

from pex import interpreter
from pex.compatibility import PY3
from pex.testing import (
    IS_PYPY,
    PY27,
    PY35,
    ensure_python_distribution,
    ensure_python_interpreter,
    temporary_dir
)

try:
  from mock import patch
except ImportError:
  from unittest.mock import patch


def tuple_from_version(version_string):
  return tuple(int(component) for component in version_string.split('.'))


class TestPythonInterpreter(object):

  @pytest.mark.skipif(PY3,
                      reason='This test relies on the `reload` builtin which is not available in '
                             'python 3 as a builtin.')
  def test_all_does_not_raise_with_empty_path_envvar(self):
    """ additionally, tests that the module does not raise at import """
    with patch.dict(os.environ, clear=True):
      reload(interpreter)
      interpreter.PythonInterpreter.all()

  TEST_INTERPRETER1_VERSION = PY27
  TEST_INTERPRETER1_VERSION_TUPLE = tuple_from_version(TEST_INTERPRETER1_VERSION)

  TEST_INTERPRETER2_VERSION = PY35
  TEST_INTERPRETER2_VERSION_TUPLE = tuple_from_version(TEST_INTERPRETER2_VERSION)

  @pytest.fixture
  def test_interpreter1(self):
    return ensure_python_interpreter(self.TEST_INTERPRETER1_VERSION)

  @pytest.fixture
  def test_interpreter2(self):
    return ensure_python_interpreter(self.TEST_INTERPRETER2_VERSION)

  @pytest.mark.skipif(IS_PYPY)
  def test_interpreter_versioning(self, test_interpreter1):
    py_interpreter = interpreter.PythonInterpreter.from_binary(test_interpreter1)
    assert py_interpreter.identity.version == self.TEST_INTERPRETER1_VERSION_TUPLE

  @pytest.mark.skipif(IS_PYPY)
  def test_interpreter_caching_basic(self, test_interpreter1, test_interpreter2):
    py_interpreter1 = interpreter.PythonInterpreter.from_binary(test_interpreter1)
    py_interpreter2 = interpreter.PythonInterpreter.from_binary(test_interpreter2)
    assert py_interpreter1 is not py_interpreter2
    assert py_interpreter2.identity.version == self.TEST_INTERPRETER2_VERSION_TUPLE

    py_interpreter3 = interpreter.PythonInterpreter.from_binary(test_interpreter1)
    assert py_interpreter1 is py_interpreter3

  @pytest.mark.skipif(IS_PYPY)
  def test_interpreter_caching_include_site_extras(self, test_interpreter1):
    py_interpreter1 = interpreter.PythonInterpreter.from_binary(test_interpreter1,
                                                                include_site_extras=False)
    py_interpreter2 = interpreter.PythonInterpreter.from_binary(test_interpreter1,
                                                                include_site_extras=True)
    py_interpreter3 = interpreter.PythonInterpreter.from_binary(test_interpreter1)
    assert py_interpreter1 is not py_interpreter2
    assert py_interpreter1.identity.version == py_interpreter2.identity.version
    assert py_interpreter2 is py_interpreter3

  @pytest.mark.skipif(IS_PYPY)
  def test_interpreter_caching_path_extras(self):
    python, pip = ensure_python_distribution(self.TEST_INTERPRETER1_VERSION)
    with temporary_dir() as td:
      path_extra = os.path.realpath(td)
      subprocess.check_call([pip,
                             'install',
                             '--target={}'.format(path_extra),
                             'ansicolors==1.1.8'])
      py_interpreter1 = interpreter.PythonInterpreter.from_binary(python,
                                                                  path_extras=[path_extra],
                                                                  include_site_extras=False)
      py_interpreter2 = interpreter.PythonInterpreter.from_binary(python,
                                                                  include_site_extras=False)
      py_interpreter3 = interpreter.PythonInterpreter.from_binary(python,
                                                                  path_extras=[path_extra],
                                                                  include_site_extras=False)
      assert py_interpreter1 is not py_interpreter2
      assert py_interpreter1.extras == {('ansicolors', '1.1.8'): path_extra}
      assert py_interpreter2.extras == {}
      assert py_interpreter1 is py_interpreter3



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import subprocess
import sys
import textwrap
from contextlib import contextmanager
from types import ModuleType

import pytest
from twitter.common.contextutil import temporary_file

from pex.bin.pex import setup_interpreter
from pex.compatibility import PY2, WINDOWS, nested, to_bytes
from pex.installer import EggInstaller, WheelInstaller
from pex.interpreter import PythonInterpreter
from pex.pex import PEX
from pex.pex_builder import PEXBuilder
from pex.pex_info import PexInfo
from pex.resolver import resolve
from pex.testing import (
    IS_PYPY,
    PY27,
    PY36,
    ensure_python_interpreter,
    make_bdist,
    make_installer,
    named_temporary_file,
    run_simple_pex_test,
    temporary_dir,
    write_simple_pex
)
from pex.util import DistributionHelper

try:
  from unittest import mock
except ImportError:
  import mock


@pytest.mark.skipif('sys.version_info > (3,)')
def test_pex_uncaught_exceptions():
  body = "raise Exception('This is an exception')"
  so, rc = run_simple_pex_test(body)
  assert b'This is an exception' in so, 'Standard out was: %s' % so
  assert rc == 1


def test_excepthook_honored():
  body = textwrap.dedent("""
  import sys

  def excepthook(ex_type, ex, tb):
    print('Custom hook called with: {0}'.format(ex))
    sys.exit(42)

  sys.excepthook = excepthook

  raise Exception('This is an exception')
  """)

  so, rc = run_simple_pex_test(body)
  assert so == b'Custom hook called with: This is an exception\n', 'Standard out was: %s' % so
  assert rc == 42


def _test_sys_exit(arg, expected_output, expected_rc):
  body = "import sys; sys.exit({arg})".format(arg=arg)
  so, rc = run_simple_pex_test(body)
  assert so == expected_output, 'Should not print SystemExit traceback.'
  assert rc == expected_rc


def test_pex_sys_exit_does_not_print_for_numeric_value():
  _test_sys_exit(2, b'', 2)


def test_pex_sys_exit_prints_non_numeric_value_no_traceback():
  text = 'something went wrong'

  sys_exit_arg = '"' + text + '"'
  # encode the string somehow that's compatible with 2 and 3
  expected_output = to_bytes(text) + b'\n'
  _test_sys_exit(sys_exit_arg, expected_output, 1)


def test_pex_sys_exit_doesnt_print_none():
  _test_sys_exit('', to_bytes(''), 0)


def test_pex_sys_exit_prints_objects():
  _test_sys_exit('Exception("derp")', to_bytes('derp\n'), 1)


@pytest.mark.skipif('hasattr(sys, "pypy_version_info")')
def test_pex_atexit_swallowing():
  body = textwrap.dedent("""
  import atexit

  def raise_on_exit():
    raise Exception('This is an exception')

  atexit.register(raise_on_exit)
  """)

  so, rc = run_simple_pex_test(body)
  assert so == b''
  assert rc == 0

  env_copy = os.environ.copy()
  env_copy.update(PEX_TEARDOWN_VERBOSE='1')
  so, rc = run_simple_pex_test(body, env=env_copy)
  assert b'This is an exception' in so
  assert rc == 0


def test_minimum_sys_modules():
  # builtins stay
  builtin_module = ModuleType('my_builtin')
  modules = {'my_builtin': builtin_module}
  new_modules = PEX.minimum_sys_modules([], modules)
  assert new_modules == modules
  new_modules = PEX.minimum_sys_modules(['bad_path'], modules)
  assert new_modules == modules

  # tainted evict
  tainted_module = ModuleType('tainted_module')
  tainted_module.__path__ = ['bad_path']
  modules = {'tainted_module': tainted_module}
  new_modules = PEX.minimum_sys_modules([], modules)
  assert new_modules == modules
  new_modules = PEX.minimum_sys_modules(['bad_path'], modules)
  assert new_modules == {}
  assert tainted_module.__path__ == []

  # tainted cleaned
  tainted_module = ModuleType('tainted_module')
  tainted_module.__path__ = ['bad_path', 'good_path']
  modules = {'tainted_module': tainted_module}
  new_modules = PEX.minimum_sys_modules([], modules)
  assert new_modules == modules
  new_modules = PEX.minimum_sys_modules(['bad_path'], modules)
  assert new_modules == modules
  assert tainted_module.__path__ == ['good_path']

  # If __path__ is not a list the module is removed; typically this implies
  # it's a namespace package (https://www.python.org/dev/peps/pep-0420/) where
  # __path__ is a _NamespacePath.
  try:
    from importlib._bootstrap_external import _NamespacePath
    bad_path = _NamespacePath("hello", "world", None)
  except ImportError:
    bad_path = {"hello": "world"}
  class FakeModule(object):
    pass
  tainted_module = FakeModule()
  tainted_module.__path__ = bad_path   # Not a list as expected
  modules = {'tainted_module': tainted_module}
  new_modules = PEX.minimum_sys_modules(['bad_path'], modules)
  assert new_modules == {}


def test_site_libs():
  with nested(mock.patch.object(PEX, '_get_site_packages'), temporary_dir()) as (
          mock_site_packages, tempdir):
    site_packages = os.path.join(tempdir, 'site-packages')
    os.mkdir(site_packages)
    mock_site_packages.return_value = set([site_packages])
    site_libs = PEX.site_libs()
    assert site_packages in site_libs


@pytest.mark.skipif(WINDOWS, reason='No symlinks on windows')
def test_site_libs_symlink():
  with nested(mock.patch.object(PEX, '_get_site_packages'), temporary_dir()) as (
          mock_site_packages, tempdir):
    site_packages = os.path.join(tempdir, 'site-packages')
    os.mkdir(site_packages)
    site_packages_link = os.path.join(tempdir, 'site-packages-link')
    os.symlink(site_packages, site_packages_link)
    mock_site_packages.return_value = set([site_packages_link])

    site_libs = PEX.site_libs()
    assert os.path.realpath(site_packages) in site_libs
    assert site_packages_link in site_libs


def test_site_libs_excludes_prefix():
  """Windows returns sys.prefix as part of getsitepackages(). Make sure to exclude it."""

  with nested(mock.patch.object(PEX, '_get_site_packages'), temporary_dir()) as (
          mock_site_packages, tempdir):
    site_packages = os.path.join(tempdir, 'site-packages')
    os.mkdir(site_packages)
    mock_site_packages.return_value = set([site_packages, sys.prefix])
    site_libs = PEX.site_libs()
    assert site_packages in site_libs
    assert sys.prefix not in site_libs


@pytest.mark.parametrize('zip_safe', (False, True))
@pytest.mark.parametrize('project_name', ('my_project', 'my-project'))
@pytest.mark.parametrize('installer_impl', (EggInstaller, WheelInstaller))
def test_pex_script(installer_impl, project_name, zip_safe):
  kw = dict(name=project_name, installer_impl=installer_impl, zip_safe=zip_safe)
  with make_installer(**kw) as installer:
    bdist = DistributionHelper.distribution_from_path(installer.bdist())

    env_copy = os.environ.copy()
    env_copy['PEX_SCRIPT'] = 'hello_world'
    so, rc = run_simple_pex_test('', env=env_copy)
    assert rc == 1, so.decode('utf-8')
    assert b"Could not find script 'hello_world'" in so

    so, rc = run_simple_pex_test('', env=env_copy, dists=[bdist])
    assert rc == 0, so.decode('utf-8')
    assert b'hello world' in so

    env_copy['PEX_SCRIPT'] = 'shell_script'
    so, rc = run_simple_pex_test('', env=env_copy, dists=[bdist])
    assert rc == 1, so.decode('utf-8')
    assert b'Unable to parse' in so


def test_pex_run():
  with named_temporary_file() as fake_stdout:
    with temporary_dir() as temp_dir:
      pex = write_simple_pex(
        temp_dir,
        'import sys; sys.stdout.write("hello"); sys.stderr.write("hello"); sys.exit(0)'
      )
      rc = PEX(pex.path()).run(stdin=None, stdout=fake_stdout, stderr=fake_stdout)
      assert rc == 0

      fake_stdout.seek(0)
      assert fake_stdout.read() == b'hellohello'


def test_pex_paths():
  # Tests that PEX_PATH allows importing sources from the referenced pex.
  with named_temporary_file() as fake_stdout:
    with temporary_dir() as temp_dir:
      pex1_path = os.path.join(temp_dir, 'pex1')
      write_simple_pex(
        pex1_path,
        exe_contents='',
        sources=[
          ('foo_pkg/__init__.py', ''),
          ('foo_pkg/foo_module.py', 'def foo_func():\n  return "42"')
        ]
      )

      pex2_path = os.path.join(temp_dir, 'pex2')
      pex2 = write_simple_pex(
        pex2_path,
        'import sys; from bar_pkg.bar_module import bar_func; '
        'sys.stdout.write(bar_func()); sys.exit(0)',
        sources=[
          ('bar_pkg/bar_module.py',
           'from foo_pkg.foo_module import foo_func\ndef bar_func():\n  return foo_func()')
        ]
      )

      rc = PEX(pex2.path()).run(stdin=None, stdout=fake_stdout, env={'PEX_PATH': pex1_path})
      assert rc == 0

      fake_stdout.seek(0)
      assert fake_stdout.read() == b'42'


@contextmanager
def _add_test_hello_to_pex(ep):
  with temporary_dir() as td:
    hello_file = "\n".join([
      "def hello():",
      "  print('hello')",
    ])
    with temporary_file(root_dir=td) as tf:
      with open(tf.name, 'w') as handle:
        handle.write(hello_file)

      pex_builder = PEXBuilder()
      pex_builder.add_source(tf.name, 'test.py')
      pex_builder.set_entry_point(ep)
      pex_builder.freeze()
      yield pex_builder


def test_pex_verify_entry_point_method_should_pass():
  with _add_test_hello_to_pex('test:hello') as pex_builder:
    # No error should happen here because `test:hello` is correct
    PEX(pex_builder.path(),
        interpreter=pex_builder.interpreter,
        verify_entry_point=True)


def test_pex_verify_entry_point_module_should_pass():
  with _add_test_hello_to_pex('test') as pex_builder:
    # No error should happen here because `test` is correct
    PEX(pex_builder.path(),
        interpreter=pex_builder.interpreter,
        verify_entry_point=True)


def test_pex_verify_entry_point_method_should_fail():
  with _add_test_hello_to_pex('test:invalid_entry_point') as pex_builder:
    # Expect InvalidEntryPoint due to invalid entry point method
    with pytest.raises(PEX.InvalidEntryPoint):
      PEX(pex_builder.path(),
          interpreter=pex_builder.interpreter,
          verify_entry_point=True)


def test_pex_verify_entry_point_module_should_fail():
  with _add_test_hello_to_pex('invalid.module') as pex_builder:
    # Expect InvalidEntryPoint due to invalid entry point module
    with pytest.raises(PEX.InvalidEntryPoint):
      PEX(pex_builder.path(),
          interpreter=pex_builder.interpreter,
          verify_entry_point=True)


@pytest.mark.skipif(IS_PYPY)
def test_activate_interpreter_different_from_current():
  with temporary_dir() as pex_root:
    interp_version = PY36 if PY2 else PY27
    custom_interpreter = setup_interpreter(
      interpreter=PythonInterpreter.from_binary(ensure_python_interpreter(interp_version)),
      interpreter_cache_dir=os.path.join(pex_root, 'interpreters'),
      repos=None,  # Default to PyPI.
      use_wheel=True
    )
    pex_info = PexInfo.default(custom_interpreter)
    pex_info.pex_root = pex_root
    with temporary_dir() as pex_chroot:
      pex_builder = PEXBuilder(path=pex_chroot,
                               interpreter=custom_interpreter,
                               pex_info=pex_info)
      with make_bdist(installer_impl=WheelInstaller, interpreter=custom_interpreter) as bdist:
        pex_builder.add_distribution(bdist)
        pex_builder.set_entry_point('sys:exit')
        pex_builder.freeze()

        pex = PEX(pex_builder.path(), interpreter=custom_interpreter)
        try:
          pex._activate()
        except SystemExit as e:
          pytest.fail('PEX activation of %s failed with %s' % (pex, e))


def test_execute_interpreter_dashc_program():
  with temporary_dir() as pex_chroot:
    pex_builder = PEXBuilder(path=pex_chroot)
    pex_builder.freeze()
    process = PEX(pex_chroot).run(args=['-c', 'import sys; print(" ".join(sys.argv))', 'one'],
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE,
                                  blocking=False)
    stdout, stderr = process.communicate()

    assert 0 == process.returncode
    assert b'-c one\n' == stdout
    assert b'' == stderr


def test_execute_interpreter_dashm_module():
  with temporary_dir() as pex_chroot:
    pex_builder = PEXBuilder(path=pex_chroot)
    with temporary_file(root_dir=pex_chroot) as fp:
      fp.write(b'import sys; print(" ".join(sys.argv))')
      fp.close()
      pex_builder.add_source(fp.name, 'foo/bar.py')
    pex_builder.freeze()
    process = PEX(pex_chroot).run(args=['-m', 'foo.bar', 'one', 'two'],
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE,
                                  blocking=False)
    stdout, stderr = process.communicate()

    assert 0 == process.returncode
    assert b'foo.bar one two\n' == stdout
    assert b'' == stderr


def test_execute_interpreter_stdin_program():
  with temporary_dir() as pex_chroot:
    pex_builder = PEXBuilder(path=pex_chroot)
    pex_builder.freeze()
    process = PEX(pex_chroot).run(args=['-', 'one', 'two'],
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE,
                                  stdin=subprocess.PIPE,
                                  blocking=False)
    stdout, stderr = process.communicate(input=b'import sys; print(" ".join(sys.argv))')

    assert 0 == process.returncode
    assert b'- one two\n' == stdout
    assert b'' == stderr


def test_execute_interpreter_file_program():
  with temporary_dir() as pex_chroot:
    pex_builder = PEXBuilder(path=pex_chroot)
    pex_builder.freeze()
    with temporary_file(root_dir=pex_chroot) as fp:
      fp.write(b'import sys; print(" ".join(sys.argv))')
      fp.close()
      process = PEX(pex_chroot).run(args=[fp.name, 'one', 'two'],
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    blocking=False)
      stdout, stderr = process.communicate()

      assert 0 == process.returncode
      assert '{} one two\n'.format(fp.name).encode('utf-8') == stdout
      assert b'' == stderr


def test_pex_run_custom_setuptools_useable():
  with temporary_dir() as resolve_cache:
    dists = [resolved_dist.distribution
             for resolved_dist in resolve(['setuptools==36.2.7'], cache=resolve_cache)]
    with temporary_dir() as temp_dir:
      pex = write_simple_pex(
        temp_dir,
        'from setuptools.sandbox import run_setup',
        dists=dists,
      )
      rc = PEX(pex.path()).run()
      assert rc == 0


def test_pex_run_conflicting_custom_setuptools_useable():
  # Here we use an older setuptools to build the pex which has a newer setuptools requirement.
  # These setuptools dists have different pkg_resources APIs:
  # $ diff \
  #   <(zipinfo -1 setuptools-20.3.1-py2.py3-none-any.whl | grep pkg_resources/ | sort) \
  #   <(zipinfo -1 setuptools-40.4.3-py2.py3-none-any.whl | grep pkg_resources/ | sort)
  # 2a3,4
  # > pkg_resources/py31compat.py
  # > pkg_resources/_vendor/appdirs.py
  with temporary_dir() as resolve_cache:
    dists = [resolved_dist.distribution
             for resolved_dist in resolve(['setuptools==20.3.1'], cache=resolve_cache)]
    interpreter = PythonInterpreter.from_binary(sys.executable,
                                                path_extras=[dist.location for dist in dists],
                                                include_site_extras=False)
    dists = [resolved_dist.distribution
             for resolved_dist in resolve(['setuptools==40.4.3'], cache=resolve_cache)]
    with temporary_dir() as temp_dir:
      pex = write_simple_pex(
        temp_dir,
        'from pkg_resources import appdirs, py31compat',
        dists=dists,
        interpreter=interpreter
      )
      rc = PEX(pex.path()).run()
      assert rc == 0



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import zipimport

import pkg_resources
import pytest

from pex.compatibility import to_bytes
from pex.finders import ChainedFinder
from pex.finders import _add_finder as add_finder
from pex.finders import _remove_finder as remove_finder
from pex.finders import (
    find_eggs_in_zip,
    find_wheels_in_zip,
    get_entry_point_from_console_script,
    get_script_from_egg,
    get_script_from_whl
)

try:
  import mock
except ImportError:
  from unittest import mock


def test_chained_finder():
  def finder1(importer, path_item, only=False):
    for foo in ('foo', 'bar'):
      yield foo

  def finder2(importer, path_item, only=False):
    yield 'baz'

  cf = ChainedFinder([finder1])
  assert list(cf(None, None)) == ['foo', 'bar']

  cf = ChainedFinder([finder1, finder2])
  assert list(cf(None, None)) == ['foo', 'bar', 'baz']


GET_FINDER = 'pex.finders._get_finder'
REGISTER_FINDER = 'pex.finders.pkg_resources.register_finder'


def test_add_new_finder():
  with mock.patch(GET_FINDER) as mock_get_finder:
    with mock.patch(REGISTER_FINDER) as mock_register_finder:
      mock_get_finder.return_value = None
      add_finder('foo', 'bar')
      mock_register_finder.assert_called_with('foo', 'bar')


def test_append_finder():
  with mock.patch(GET_FINDER) as mock_get_finder:
    with mock.patch(REGISTER_FINDER) as mock_register_finder:
      mock_get_finder.return_value = 'bar'
      add_finder('foo', 'baz')
      mock_register_finder.assert_called_with('foo', ChainedFinder(['bar', 'baz']))

  with mock.patch(GET_FINDER) as mock_get_finder:
    with mock.patch(REGISTER_FINDER) as mock_register_finder:
      mock_get_finder.return_value = ChainedFinder(['bar'])
      add_finder('foo', 'baz')
      mock_register_finder.assert_called_with('foo', ChainedFinder(['bar', 'baz']))


def test_remove_finder():
  # wasn't registered
  with mock.patch(GET_FINDER) as mock_get_finder:
    with mock.patch(REGISTER_FINDER) as mock_register_finder:
      mock_get_finder.return_value = None
      remove_finder('foo', 'baz')
      assert not mock_register_finder.called

  # was registered but we're asking for the wrong one
  with mock.patch(GET_FINDER) as mock_get_finder:
    with mock.patch(REGISTER_FINDER) as mock_register_finder:
      mock_get_finder.return_value = ChainedFinder(['bar'])
      remove_finder('foo', 'baz')
      assert not mock_register_finder.called

  # was registered but we're asking for the wrong one
  with mock.patch(GET_FINDER) as mock_get_finder:
    with mock.patch(REGISTER_FINDER) as mock_register_finder:
      cf = ChainedFinder(['bar', 'baz', 'bak'])
      mock_get_finder.return_value = cf
      remove_finder('foo', 'baz')
      assert cf.finders == ['bar', 'bak']
      assert not mock_register_finder.called

  # was registered but we're asking for the wrong one
  with mock.patch(GET_FINDER) as mock_get_finder:
    with mock.patch(REGISTER_FINDER) as mock_register_finder:
      cf = ChainedFinder(['bar', 'baz'])
      mock_get_finder.return_value = cf
      remove_finder('foo', 'baz')
      mock_register_finder.assert_called_with('foo', 'bar')

  # was registered but we're asking for the wrong one
  with mock.patch(GET_FINDER) as mock_get_finder:
    with mock.patch(REGISTER_FINDER) as mock_register_finder:
      mock_get_finder.return_value = 'bar'
      remove_finder('foo', 'bar')
      mock_register_finder.assert_called_with('foo', pkg_resources.find_nothing)

  # was registered but we're asking for the wrong one
  with mock.patch(GET_FINDER) as mock_get_finder:
    with mock.patch(REGISTER_FINDER) as mock_register_finder:
      mock_get_finder.return_value = ChainedFinder(['bar'])
      remove_finder('foo', 'bar')
      mock_register_finder.assert_called_with('foo', pkg_resources.find_nothing)


def test_get_script_from_egg_with_no_scripts():
  # Make sure eggs without scripts don't cause errors.
  egg_path = './tests/example_packages/Flask_Cache-0.13.1-py2.7.egg'
  dists = list(find_eggs_in_zip(zipimport.zipimporter(egg_path), egg_path, only=True))
  assert len(dists) == 1

  dist = dists[0]
  assert (None, None) == get_script_from_egg('non_existent_script', dist)


def test_get_script_from_egg():
  egg_path = './tests/example_packages/eno-0.0.17-py2.7.egg'
  dists = list(find_eggs_in_zip(zipimport.zipimporter(egg_path), egg_path, only=True))
  assert len(dists) == 1

  dist = dists[0]

  location, content = get_script_from_egg('run_eno_server', dist)
  assert os.path.join(egg_path, 'EGG-INFO/scripts/run_eno_server') == location
  assert content.startswith('#!'), 'Expected a `scripts` style script with shebang.'

  assert (None, None) == get_script_from_egg('non_existent_script', dist)


# In-part, tests a bug where the wheel distribution name has dashes as reported in:
#   https://github.com/pantsbuild/pex/issues/443
#   https://github.com/pantsbuild/pex/issues/551
def test_get_script_from_whl():
  whl_path = './tests/example_packages/aws_cfn_bootstrap-1.4-py2-none-any.whl'
  dists = list(find_wheels_in_zip(zipimport.zipimporter(whl_path), whl_path))
  assert len(dists) == 1

  dist = dists[0]
  assert 'aws-cfn-bootstrap' == dist.project_name

  script_path, script_content = get_script_from_whl('cfn-signal', dist)
  assert os.path.join(whl_path, 'aws_cfn_bootstrap-1.4.data/scripts/cfn-signal') == script_path
  assert script_content.startswith(to_bytes('#!')), 'Expected a `scripts`-style script w/shebang.'

  assert (None, None) == get_script_from_whl('non_existent_script', dist)


class FakeDist(object):
  def __init__(self, key, console_script_entry):
    self.key = key
    script = console_script_entry.split('=')[0].strip()
    self._entry_map = {'console_scripts': {script: console_script_entry}}

  def get_entry_map(self):
    return self._entry_map


def test_get_entry_point_from_console_script():
  dists = [FakeDist(key='fake', console_script_entry='bob= bob.main:run'),
           FakeDist(key='fake', console_script_entry='bob =bob.main:run')]

  dist, entrypoint = get_entry_point_from_console_script('bob', dists)
  assert 'bob.main:run' == entrypoint
  assert dist in dists


def test_get_entry_point_from_console_script_conflict():
  dists = [FakeDist(key='bob', console_script_entry='bob= bob.main:run'),
           FakeDist(key='fake', console_script_entry='bob =bob.main:run')]
  with pytest.raises(RuntimeError):
    get_entry_point_from_console_script('bob', dists)


def test_get_entry_point_from_console_script_dne():
  dists = [FakeDist(key='bob', console_script_entry='bob= bob.main:run'),
           FakeDist(key='fake', console_script_entry='bob =bob.main:run')]
  assert (None, None) == get_entry_point_from_console_script('jane', dists)



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import pytest

from pex.compatibility import to_bytes, to_unicode, unicode_string


def test_to_bytes():
  assert isinstance(to_bytes(''), bytes)
  assert isinstance(to_bytes('abc'), bytes)
  assert isinstance(to_bytes(b'abc'), bytes)
  assert isinstance(to_bytes(u'abc'), bytes)
  assert isinstance(to_bytes(b'abc'.decode('latin-1'), encoding='utf-8'), bytes)

  for bad_value in (123, None):
    with pytest.raises(ValueError):
      to_bytes(bad_value)


def test_to_unicode():
  assert isinstance(to_unicode(''), unicode_string)
  assert isinstance(to_unicode('abc'), unicode_string)
  assert isinstance(to_unicode(b'abc'), unicode_string)
  assert isinstance(to_unicode(u'abc'), unicode_string)
  assert isinstance(to_unicode(u'abc'.encode('latin-1'), encoding='latin-1'), unicode_string)

  for bad_value in (123, None):
    with pytest.raises(ValueError):
      to_unicode(bad_value)



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import time

import pytest
from twitter.common.contextutil import temporary_dir

from pex.common import safe_copy
from pex.crawler import Crawler
from pex.fetcher import Fetcher
from pex.package import EggPackage, SourcePackage
from pex.resolvable import ResolvableRequirement
from pex.resolver import Resolver, Unsatisfiable, _ResolvableSet, resolve_multi
from pex.resolver_options import ResolverOptionsBuilder
from pex.testing import make_sdist


def test_empty_resolve():
  empty_resolve_multi = list(resolve_multi([]))
  assert empty_resolve_multi == []

  with temporary_dir() as td:
    empty_resolve_multi = list(resolve_multi([], cache=td))
    assert empty_resolve_multi == []


def test_simple_local_resolve():
  project_sdist = make_sdist(name='project')

  with temporary_dir() as td:
    safe_copy(project_sdist, os.path.join(td, os.path.basename(project_sdist)))
    fetchers = [Fetcher([td])]
    resolved_dists = list(resolve_multi(['project'], fetchers=fetchers))
    assert len(resolved_dists) == 1


def test_diamond_local_resolve_cached():
  # This exercises the issue described here: https://github.com/pantsbuild/pex/issues/120
  project1_sdist = make_sdist(name='project1', install_reqs=['project2<1.0.0'])
  project2_sdist = make_sdist(name='project2')

  with temporary_dir() as dd:
    for sdist in (project1_sdist, project2_sdist):
      safe_copy(sdist, os.path.join(dd, os.path.basename(sdist)))
    fetchers = [Fetcher([dd])]
    with temporary_dir() as cd:
      resolved_dists = list(
        resolve_multi(['project1', 'project2'], fetchers=fetchers, cache=cd, cache_ttl=1000)
      )
      assert len(resolved_dists) == 2


def test_cached_dependency_pinned_unpinned_resolution_multi_run():
  # This exercises the issue described here: https://github.com/pantsbuild/pex/issues/178
  project1_0_0 = make_sdist(name='project', version='1.0.0')
  project1_1_0 = make_sdist(name='project', version='1.1.0')

  with temporary_dir() as td:
    for sdist in (project1_0_0, project1_1_0):
      safe_copy(sdist, os.path.join(td, os.path.basename(sdist)))
    fetchers = [Fetcher([td])]
    with temporary_dir() as cd:
      # First run, pinning 1.0.0 in the cache
      resolved_dists = list(
        resolve_multi(['project', 'project==1.0.0'],
                      fetchers=fetchers,
                      cache=cd,
                      cache_ttl=1000)
      )
      assert len(resolved_dists) == 1
      assert resolved_dists[0].distribution.version == '1.0.0'
      # This simulates separate invocations of pex but allows us to keep the same tmp cache dir
      Crawler.reset_cache()
      # Second, run, the unbounded 'project' req will find the 1.0.0 in the cache. But should also
      # return SourcePackages found in td
      resolved_dists = list(
        resolve_multi(['project', 'project==1.1.0'],
                      fetchers=fetchers,
                      cache=cd,
                      cache_ttl=1000)
      )
      assert len(resolved_dists) == 1
      assert resolved_dists[0].distribution.version == '1.1.0'
      # Third run, if exact resolvable and inexact resolvable, and cache_ttl is expired, exact
      # resolvable should pull from pypi as well since inexact will and the resulting
      # resolvable_set.merge() would fail.
      Crawler.reset_cache()
      time.sleep(1)
      resolved_dists = list(
        resolve_multi(['project', 'project==1.1.0'],
                      fetchers=fetchers,
                      cache=cd,
                      cache_ttl=1)
      )
      assert len(resolved_dists) == 1
      assert resolved_dists[0].distribution.version == '1.1.0'


def test_ambiguous_transitive_resolvable():
  # If an unbounded or larger bounded resolvable is resolved first, and a
  # transitive resolvable is resolved later in another round, Error(Ambiguous resolvable) can be
  # raised because foo pulls in foo-2.0.0 and bar->foo==1.0.0 pulls in foo-1.0.0.
  foo1_0 = make_sdist(name='foo', version='1.0.0')
  foo2_0 = make_sdist(name='foo', version='2.0.0')
  bar1_0 = make_sdist(name='bar', version='1.0.0', install_reqs=['foo==1.0.0'])
  with temporary_dir() as td:
    for sdist in (foo1_0, foo2_0, bar1_0):
      safe_copy(sdist, os.path.join(td, os.path.basename(sdist)))
    fetchers = [Fetcher([td])]
    with temporary_dir() as cd:
      resolved_dists = list(
        resolve_multi(['foo', 'bar'],
                      fetchers=fetchers,
                      cache=cd,
                      cache_ttl=1000)
      )
      assert len(resolved_dists) == 2
      assert resolved_dists[0].distribution.version == '1.0.0'


def test_resolve_prereleases():
  stable_dep = make_sdist(name='dep', version='2.0.0')
  prerelease_dep = make_sdist(name='dep', version='3.0.0rc3')

  with temporary_dir() as td:
    for sdist in (stable_dep, prerelease_dep):
      safe_copy(sdist, os.path.join(td, os.path.basename(sdist)))
    fetchers = [Fetcher([td])]

    def assert_resolve(expected_version, **resolve_kwargs):
      resolved_dists = list(
        resolve_multi(['dep>=1,<4'], fetchers=fetchers, **resolve_kwargs)
      )
      assert 1 == len(resolved_dists)
      resolved_dist = resolved_dists[0]
      assert expected_version == resolved_dist.distribution.version

    assert_resolve('2.0.0')
    assert_resolve('2.0.0', allow_prereleases=False)
    assert_resolve('3.0.0rc3', allow_prereleases=True)


def test_resolve_prereleases_cached():
  stable_dep = make_sdist(name='dep', version='2.0.0')
  prerelease_dep = make_sdist(name='dep', version='3.0.0rc3')

  with temporary_dir() as td:
    for sdist in (stable_dep, prerelease_dep):
      safe_copy(sdist, os.path.join(td, os.path.basename(sdist)))
    fetchers = [Fetcher([td])]

    with temporary_dir() as cd:
      def assert_resolve(dep, expected_version, **resolve_kwargs):
        resolved_dists = list(
          resolve_multi([dep], cache=cd, cache_ttl=1000, **resolve_kwargs)
        )
        assert 1 == len(resolved_dists)
        resolved_dist = resolved_dists[0]
        assert expected_version == resolved_dist.distribution.version

      Crawler.reset_cache()

      # First do a run to load it into the cache.
      assert_resolve('dep>=1,<4', '3.0.0rc3', allow_prereleases=True, fetchers=fetchers)

      # This simulates running from another pex command. The Crawler cache actually caches an empty
      # cache so this fails in the same "process".
      Crawler.reset_cache()

      # Now assert that we can get it from the cache by removing the source.
      assert_resolve('dep>=1,<4', '3.0.0rc3', allow_prereleases=True, fetchers=[])

      # It should also be able to resolve without allow_prereleases, if explicitly requested.
      Crawler.reset_cache()
      assert_resolve('dep>=1.rc1,<4', '3.0.0rc3', fetchers=[])


def test_resolve_prereleases_and_no_version():
  prerelease_dep = make_sdist(name='dep', version='3.0.0rc3')

  with temporary_dir() as td:
    safe_copy(prerelease_dep, os.path.join(td, os.path.basename(prerelease_dep)))
    fetchers = [Fetcher([td])]

    def assert_resolve(deps, expected_version, **resolve_kwargs):
      resolved_dists = list(
        resolve_multi(deps, fetchers=fetchers, **resolve_kwargs)
      )
      assert 1 == len(resolved_dists)
      resolved_dist = resolved_dists[0]
      assert expected_version == resolved_dist.distribution.version

    # When allow_prereleases is specified, the requirement (from two dependencies)
    # for a specific pre-release version and no version specified, accepts the pre-release
    # version correctly.
    assert_resolve(['dep==3.0.0rc3', 'dep'], '3.0.0rc3', allow_prereleases=True)

    # Without allow_prereleases set, the pre-release version is rejected.
    # This used to be an issue when a command-line use did not pass the `--pre` option
    # correctly into the API call for resolve_multi() from build_pex() in pex.py.
    with pytest.raises(Unsatisfiable):
      assert_resolve(['dep==3.0.0rc3', 'dep'], '3.0.0rc3')


def test_resolve_prereleases_multiple_set():
  stable_dep = make_sdist(name='dep', version='2.0.0')
  prerelease_dep1 = make_sdist(name='dep', version='3.0.0rc3')
  prerelease_dep2 = make_sdist(name='dep', version='3.0.0rc4')
  prerelease_dep3 = make_sdist(name='dep', version='3.0.0rc5')

  with temporary_dir() as td:
    for sdist in (stable_dep, prerelease_dep1, prerelease_dep2, prerelease_dep3):
      safe_copy(sdist, os.path.join(td, os.path.basename(sdist)))
    fetchers = [Fetcher([td])]

    def assert_resolve(expected_version, **resolve_kwargs):
      resolved_dists = list(
        resolve_multi(['dep>=3.0.0rc1', 'dep==3.0.0rc4'],
                      fetchers=fetchers,
                      **resolve_kwargs)
      )
      assert 1 == len(resolved_dists)
      resolved_dist = resolved_dists[0]
      assert expected_version == resolved_dist.distribution.version

    # This should resolve with explicit prerelease being set or implicitly.
    assert_resolve('3.0.0rc4', allow_prereleases=True)
    assert_resolve('3.0.0rc4')


def test_resolvable_set():
  builder = ResolverOptionsBuilder()
  rs = _ResolvableSet()
  rq = ResolvableRequirement.from_string('foo[ext]', builder)
  source_pkg = SourcePackage.from_href('foo-2.3.4.tar.gz')
  binary_pkg = EggPackage.from_href('Foo-2.3.4-py3.4.egg')

  rs.merge(rq, [source_pkg, binary_pkg])
  assert rs.get(source_pkg.name) == set([source_pkg, binary_pkg])
  assert rs.get(binary_pkg.name) == set([source_pkg, binary_pkg])
  assert rs.packages() == [(rq, set([source_pkg, binary_pkg]), None, False)]

  # test methods
  assert rs.extras('foo') == set(['ext'])
  assert rs.extras('Foo') == set(['ext'])

  # test filtering
  rs.merge(rq, [source_pkg])
  assert rs.get('foo') == set([source_pkg])
  assert rs.get('Foo') == set([source_pkg])

  with pytest.raises(Unsatisfiable):
    rs.merge(rq, [binary_pkg])


def test_resolvable_set_is_constraint_only():
  builder = ResolverOptionsBuilder()
  rs = _ResolvableSet()
  c = ResolvableRequirement.from_string('foo', builder)
  c.is_constraint = True

  package = SourcePackage.from_href('foo-2.3.4.tar.gz')
  rs.merge(c, [package])

  assert rs.packages() == [(c, set([package]), None, True)]


def test_resolvable_set_constraint_and_non_constraint():
  builder = ResolverOptionsBuilder()
  rs = _ResolvableSet()
  constraint = ResolvableRequirement.from_string('foo', builder)
  constraint.is_constraint = True

  package = SourcePackage.from_href('foo-2.3.4.tar.gz')

  rq = ResolvableRequirement.from_string('foo', builder)
  rs.merge(constraint, [package])
  rs.merge(rq, [package])

  assert rs.packages() == [(rq, set([package]), None, False)]


def test_constraints_limits_versions_usable():
  builder = ResolverOptionsBuilder()
  rs = _ResolvableSet()
  req = ResolvableRequirement.from_string("foo>0.5", builder)
  constraint = ResolvableRequirement.from_string("foo==0.7", builder)
  constraint.is_constraint = True

  version_packages = []
  for version in range(6, 10):
    version_string = "foo-0.{0}.tar.gz".format(version)
    package = SourcePackage.from_href(version_string)
    version_packages.append(package)
  rs.merge(req, version_packages)
  rs.merge(constraint, [version_packages[1]])
  assert rs.packages() == [(req, set([version_packages[1]]), None, False)]


def test_constraints_range():
  builder = ResolverOptionsBuilder()
  rs = _ResolvableSet()
  req = ResolvableRequirement.from_string("foo>0.5", builder)
  constraint = ResolvableRequirement.from_string("foo<0.9", builder)
  constraint.is_constraint = True

  version_packages = []
  for version in range(1, 10):
    version_string = "foo-0.{0}.tar.gz".format(version)
    package = SourcePackage.from_href(version_string)
    version_packages.append(package)
  rs.merge(req, version_packages[4:])
  rs.merge(constraint, version_packages[:8])
  assert rs.packages() == [(req, set(version_packages[4:8]), None, False)]


def test_resolver_with_constraint():
  builder = ResolverOptionsBuilder()
  r = Resolver()
  rs = _ResolvableSet()
  constraint = ResolvableRequirement.from_string('foo', builder)
  constraint.is_constraint = True

  package = SourcePackage.from_href('foo-2.3.4.tar.gz')

  rq = ResolvableRequirement.from_string('foo', builder)
  rs.merge(constraint, [package])
  rs.merge(rq, [package])
  assert r.resolve([], resolvable_set=rs) == []


def test_resolvable_set_built():
  builder = ResolverOptionsBuilder()
  rs = _ResolvableSet()
  rq = ResolvableRequirement.from_string('foo', builder)
  source_pkg = SourcePackage.from_href('foo-2.3.4.tar.gz')
  binary_pkg = EggPackage.from_href('foo-2.3.4-py3.4.egg')

  rs.merge(rq, [source_pkg])
  assert rs.get('foo') == set([source_pkg])
  assert rs.packages() == [(rq, set([source_pkg]), None, False)]

  with pytest.raises(Unsatisfiable):
    rs.merge(rq, [binary_pkg])

  updated_rs = rs.replace_built({source_pkg: binary_pkg})
  updated_rs.merge(rq, [binary_pkg])
  assert updated_rs.get('foo') == set([binary_pkg])
  assert updated_rs.packages() == [(rq, set([binary_pkg]), None, False)]
